#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import argparse
import json
import logging
import os

from scraper import code_gov
from scraper.util import configure_logging

logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(description='Scrape code repositories for Code.gov / DOE CODE')

    parser.add_argument('--agency', type=str, nargs='?', default='', help='Agency Label, e.g. "DOE"')
    parser.add_argument('--method', type=str, nargs='?', default='', help='Method of measuring open source')
    parser.add_argument('--organization', type=str, nargs='?', default='', help='Force all repos to report a particular organzation')
    parser.add_argument('--contact-email', type=str, nargs='?', default='', help='Force all repos to report a particular contact email')

    parser.add_argument('--config', type=str, nargs='?', default='', help='Configuration File (*.json)')

    parser.add_argument('--github-gov-orgs', action='store_true', help='Use orgs from government.github.com/community')
    parser.add_argument('--skip-labor-hours', action='store_true', help='Skip calculation of labor hours, assume "0"')

    parser.add_argument('--doecode-json', type=str, nargs='?', default=None, help='Path to DOE CODE .json file')
    parser.add_argument('--doecode-url', type=str, nargs='?', default=None, help='URL to DOE CODE .json data')
    parser.add_argument('--doecode-api-key', type=str, nargs='?', default=None, help='DOE CODE API key for accessing --doecode-url')

    parser.add_argument('--output-path', type=str, nargs='?', default='', help='Output path for .json file')
    parser.add_argument('--output-filename', type=str, nargs='?', default='code.json', help='Output filename for .json file')

    parser.add_argument('--verbose', action='store_true', help='Enable verbose output')

    args = parser.parse_args()

    configure_logging(args.verbose)

    try:
        config_json = json.load(open(args.config))
    except (FileNotFoundError, json.JSONDecodeError):
        if args.config:
            raise
        config_json = {}

    # Update config based on commandline arguments
    if args.agency:
        config_json['agency'] = args.agency
    if args.method:
        config_json['method'] = args.method
    if args.organization:
        config_json['organization'] = args.organization
    if args.contact_email:
        config_json['contact_email'] = args.contact_email
    if args.output_path:
        config_json['output_path'] = args.output_path
    if args.skip_labor_hours:
        config_json['compute_labor_hours'] = False
    if args.github_gov_orgs:
        config_json['github_gov_orgs'] = True

    config_json['DOE CODE'] = {}
    config_json['DOE CODE']['json'] = args.doecode_json
    config_json['DOE CODE']['url'] = args.doecode_url
    config_json['DOE CODE']['api_key'] = args.doecode_api_key

    output_path = config_json.get('output_path', None)
    output_path = args.output_path or output_path
    logger.debug('Output Path: %s', output_path)

    if (output_path is not None and not os.path.exists(output_path)):
        raise RuntimeError('Invalid output path argument provided!  Make sure the output path exists and try again.')

    code_json = code_gov.process_config(config_json)

    code_gov.force_attributes(code_json, config_json)

    logger.info('Number of Projects: %s', len(code_json['releases']))

    output_filepath = args.output_filename

    if output_path is not None:
        output_filepath = os.path.join(output_path, output_filepath)

    with open(output_filepath, 'w', encoding='utf-8') as fp:
        logger.info('Writing output to: %s', output_filepath)
        fp.write(code_json.to_json())


if __name__ == '__main__':
    main()



import json
import logging

import requests

logger = logging.getLogger(__name__)


def process_json(filename):
    """
    Converts a DOE CODE .json file into DOE CODE projects
    Yields DOE CODE records from a DOE CODE .json file
    """

    logger.debug('Processing DOE CODE json: %s', filename)

    doecode_json = json.load(open(filename))

    for record in doecode_json['records']:
        yield record


def process_url(url, key):
    """
    Yields DOE CODE records from a DOE CODE .json URL response
    Converts a DOE CODE API .json URL response into DOE CODE projects
    """

    logger.debug('Fetching DOE CODE JSON: %s', url)

    if key is None:
        raise ValueError('DOE CODE API Key value is missing!')

    response = requests.get(url, headers={"Authorization": "Basic " + key})
    doecode_json = response.json()

    for record in doecode_json['records']:
        yield record


def process(filename=None, url=None, key=None):
    """
    Yeilds DOE CODE records based on provided input sources

    param:
        filename (str): Path to a DOE CODE .json file
        url (str): URL for a DOE CODE server json file
        key (str): API Key for connecting to DOE CODE server
    """

    if filename is not None:
        yield from process_json(filename)
    elif url and key:
        yield from process_url(url, key)



import json
import logging
import logging.config
import os
import re
import requests
import tempfile

from subprocess import Popen, PIPE, STDOUT  # nosec

logger = logging.getLogger(__name__)

EFFORT_REGEX = re.compile(r'Effort = ([\d\.]+) Person-months')


def execute(command, cwd=None):
    logger.debug('Forking command: %s', command)

    if cwd is None:
        cwd = os.getcwd()
    elif not os.path.isdir(cwd):
        raise ValueError('path does not exist: %s', cwd)

    process = Popen(
        command,
        cwd=cwd,
        stdout=PIPE,
        stderr=STDOUT,
        shell=False)  # nosec
    out, err = process.communicate()
    return str(out), str(err)


def configure_logging(verbose=False):
    DEFAULT_LOGGING = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                # 'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
                # 'format': '%(levelname)s: %(message)s'
                'format': '%(asctime)s - %(levelname)s: %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': 'INFO',
                'formatter': 'standard',
                'class': 'logging.StreamHandler',
            },
            'null': {
                'level': 'INFO',
                'formatter': 'standard',
                'class': 'logging.NullHandler',
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': 'DEBUG',
                'propagate': False,
            },
            'github3': {
                'handlers': ['null'],
                'level': 'DEBUG',
                'propagate': False,
            },
            'urllib3': {
                'handlers': ['null'],
                'level': 'DEBUG',
                'propagate': False,
            },
        }
    }

    if verbose:
        DEFAULT_LOGGING['handlers']['default']['level'] = 'DEBUG'
        # DEFAULT_LOGGING['loggers']['']['level'] = 'DEBUG'

    logging.config.dictConfig(DEFAULT_LOGGING)


def git_repo_to_sloc(url):
    """
    Given a Git repository URL, returns number of lines of code based on cloc

    Reference:
    - cloc: https://github.com/AlDanial/cloc
    - https://www.omg.org/spec/AFP/
        - Another potential way to calculation effort

    Sample cloc output:
        {
            "header": {
                "cloc_url": "github.com/AlDanial/cloc",
                "cloc_version": "1.74",
                "elapsed_seconds": 0.195950984954834,
                "n_files": 27,
                "n_lines": 2435,
                "files_per_second": 137.78956000769,
                "lines_per_second": 12426.5769858787
            },
            "C++": {
                "nFiles": 7,
                "blank": 121,
                "comment": 314,
                "code": 371
            },
            "C/C++ Header": {
                "nFiles": 8,
                "blank": 107,
                "comment": 604,
                "code": 191
            },
            "CMake": {
                "nFiles": 11,
                "blank": 49,
                "comment": 465,
                "code": 165
            },
            "Markdown": {
                "nFiles": 1,
                "blank": 18,
                "comment": 0,
                "code": 30
            },
            "SUM": {
                "blank": 295,
                "comment": 1383,
                "code": 757,
                "nFiles": 27
            }
        }
    """

    with tempfile.TemporaryDirectory() as tmp_dir:
        logger.debug('Cloning: url=%s tmp_dir=%s', url, tmp_dir)

        tmp_clone = os.path.join(tmp_dir, 'clone-dir')

        cmd = ['git', 'clone', '--depth=1', url, tmp_clone]
        execute(cmd)

        cmd = ['cloc', '--json', tmp_clone]
        out, _ = execute(cmd)

        try:
            json_start = out.find('{"header"')
            json_blob = out[json_start:].replace('\\n', '').replace('\'', '')
            cloc_json = json.loads(json_blob)
            sloc = cloc_json['SUM']['code']
        except json.decoder.JSONDecodeError:
            logger.debug('Error Decoding: url=%s, out=%s', url, out)
            sloc = 0

    logger.debug('SLOC: url=%s, sloc=%d', url, sloc)

    return sloc


def compute_labor_hours(sloc, month_hours='cocomo_book'):
    """
    Compute the labor hours, given a count of source lines of code

    The intention is to use the COCOMO II model to compute this value.

    References:
    - http://csse.usc.edu/tools/cocomoii.php
    - http://docs.python-guide.org/en/latest/scenarios/scrape/
    """
    # Calculation of hours in a month
    if month_hours == 'hours_per_year':
        # Use number of working hours in a year:
        # (40 Hours / week) * (52 weeks / year) / (12 months / year) ~= 173.33
        HOURS_PER_PERSON_MONTH = 40.0 * 52 / 12
    else:
        # Use value from COCOMO II Book (month_hours=='cocomo_book'):
        # Reference: https://dl.acm.org/citation.cfm?id=557000
        # This is the value used by the Code.gov team:
        # https://github.com/GSA/code-gov/blob/master/LABOR_HOUR_CALC.md
        HOURS_PER_PERSON_MONTH = 152.0

    cocomo_url = 'http://csse.usc.edu/tools/cocomoii.php'
    page = requests.post(cocomo_url, data={'new_size': sloc})

    try:
        person_months = float(EFFORT_REGEX.search(page.text).group(1))
    except AttributeError:
        logger.error('Unable to find Person Months in page text: sloc=%s', sloc)
        # If there is no match, and .search(..) returns None
        person_months = 0

    labor_hours = person_months * HOURS_PER_PERSON_MONTH
    logger.debug('sloc=%d labor_hours=%d', sloc, labor_hours)

    return labor_hours


def labor_hours_from_url(url):
    sum_sloc = git_repo_to_sloc(url)
    logger.info('SLOC: %d', sum_sloc)

    labor_hours = compute_labor_hours(sum_sloc)
    logger.info('labor_hours: %d', labor_hours)

    return labor_hours


def _prune_dict_null_str(dictionary):
    """
    Prune the "None" or emptry string values from dictionary items
    """
    for key, value in list(dictionary.items()):
        if value is None or str(value) == '':
            del dictionary[key]

        if isinstance(value, dict):
            dictionary[key] = _prune_dict_null_str(dictionary[key])

    return dictionary



#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import logging
import os

from vsts.vss_connection import VssConnection
from msrest.authentication import BasicAuthentication
from scraper.tfs.models import TFSProject

logger = logging.getLogger(__name__)

HARD_CODED_TOP = 10000


def get_projects_metadata(baseurl, token):
    logger.debug('Retrieving TFS Metdata.....')
    return [project for project in get_all_projects(baseurl, token)]


def create_tfs_connection(url, token):
    """
    Creates the TFS Connection Context
    """
    if token is None:
        token = os.environ.get('TFS_API_TOKEN', None)

    tfs_credentials = BasicAuthentication('', token)
    tfs_connection = VssConnection(base_url=url, creds=tfs_credentials)
    return tfs_connection


def create_tfs_project_analysis_client(url, token=None):
    """
    Create a project_analysis_client.py client for a Team Foundation Server Enterprise connection instance.
    This is helpful for understanding project languages, but currently blank for all our test conditions.

    If token is not provided, will attempt to use the TFS_API_TOKEN
    environment variable if present.
    """
    if token is None:
        token = os.environ.get('TFS_API_TOKEN', None)

    tfs_connection = create_tfs_connection(url, token)
    project_analysis_client = tfs_connection.get_client('vsts.project_analysis.v4_1.project_analysis_client.ProjectAnalysisClient')

    if project_analysis_client is None:
        msg = 'Unable to connect to TFS Enterprise (%s) with provided token.'
        raise RuntimeError(msg, url)

    return project_analysis_client


def create_tfs_core_client(url, token=None):
    """
    Create a core_client.py client for a Team Foundation Server Enterprise connection instance

    If token is not provided, will attempt to use the TFS_API_TOKEN
    environment variable if present.
    """
    if token is None:
        token = os.environ.get('TFS_API_TOKEN', None)

    tfs_connection = create_tfs_connection(url, token)
    tfs_client = tfs_connection.get_client('vsts.core.v4_1.core_client.CoreClient')

    if tfs_client is None:
        msg = 'Unable to connect to TFS Enterprise (%s) with provided token.'
        raise RuntimeError(msg, url)

    return tfs_client


def create_tfs_git_client(url, token=None):
    """
    Creates a TFS Git Client to pull Git repo info
    """
    if token is None:
        token = os.environ.get('TFS_API_TOKEN', None)

    tfs_connection = create_tfs_connection(url, token)
    tfs_git_client = tfs_connection.get_client('vsts.git.v4_1.git_client.GitClient')

    if tfs_git_client is None:
        msg = 'Unable to create TFS Git Client, failed to connect to TFS Enterprise (%s) with provided token.'
        raise RuntimeError(msg, url)

    return tfs_git_client


def create_tfs_tfvc_client(url, token=None):
    """
    Creates a TFS TFVC Client to pull TFVC repo info
    """
    if token is None:
        token = os.environ.get('TFS_API_TOKEN', None)

    tfs_connection = create_tfs_connection(url, token)
    tfs_tfvc_client = tfs_connection.get_client('vsts.tfvc.v4_1.tfvc_client.TfvcClient')

    if tfs_tfvc_client is None:
        msg = 'Unable to create TFS Git Client, failed to connect to TFS Enterprise (%s) with provided token.'
        raise RuntimeError(msg, url)

    return tfs_tfvc_client


def get_all_projects(url, token, top=HARD_CODED_TOP):
    """
    Returns a list of all projects with their collection info from the server. Currently limited functionality to only return the first 1000 projects.
    #TODO refactor to add multiple calls to api to retrieve all projects if more exist beyond top.
    """
    project_list = []
    tfs_client = create_tfs_core_client(url, token)

    collections = tfs_client.get_project_collections(top=top)

    for collection in collections:
        collection_client = create_tfs_core_client('{url}/{collection_name}'.format(url=url, collection_name=collection.name), token)

        logger.debug('Retrieving Projects for Project Collection: {collection_name}'.format(collection_name=collection.name))
        # Retrieves all projects in the project collection
        projects = collection_client.get_projects(top=HARD_CODED_TOP)
        # get_projects only gets the project references, have to call get_project_history_entries to get last update info for projects
        # Only calling this once per collection as its an expensive API call, wil refactor later if there is a better API call to use
        collection_history_list = collection_client.get_project_history_entries()
        for project in projects:

            # get_projects only gets team project ref objects,
            # have to call get_project to get the team project object which includes the TFS Web Url for the project
            logger.debug('Retrieving Team Project for Project: {project_name}'.format(project_name=project.name))
            projectInfo = collection_client.get_project(project.id, True, True)

            tfsProject = TFSProject(projectInfo, collection)

            logger.debug('Retrieving Last Updated and Created Info for Project: {project_name}'.format(project_name=project.name))
            tfsProject.projectLastUpdateInfo = get_project_last_update_time(collection_history_list, project.id)
            tfsProject.projectCreateInfo = get_project_create_time(collection_history_list, project.id)
            project_list.append(tfsProject)

    return project_list


def get_git_repos(url, token, collection, project):
    """
    Returns a list of all git repos for the supplied project within the supplied collection
    """
    git_client = create_tfs_git_client('{url}/{collection_name}'.format(url=url, collection_name=collection.name), token)
    logger.debug('Retrieving Git Repos for Project: {project_name}'.format(project_name=project.name))
    return git_client.get_repositories(project.id)


def get_tfvc_repos(url, token, collection, project):
    """
    Returns a list of all tfvc branches for the supplied project within the supplied collection
    """
    branch_list = []
    tfvc_client = create_tfs_tfvc_client('{url}/{collection_name}'.format(url=url, collection_name=collection.name), token)

    logger.debug('Retrieving Tfvc Branches for Project: {project_name}'.format(project_name=project.name))
    branches = tfvc_client.get_branches(project.id, True, True, False, True)
    if branches:
        branch_list.extend(branches)
    else:
        logger.debug('No Tfvcc Branches in Project: {project_name}'.format(project_name=project.name))

    return branch_list


def get_project_last_update_time(collection_history_list, projectId):
    sorted_history_list = sorted(collection_history_list, key=lambda x: x.last_update_time, reverse=True)
    return next((x for x in sorted_history_list if x.id == projectId))


def get_project_create_time(collection_history_list, projectId):
    sorted_history_list = sorted(collection_history_list, key=lambda x: x.last_update_time, reverse=False)
    return next((x for x in sorted_history_list if x.id == projectId))



#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import logging

from scraper.code_gov.models import Metadata, Project
from scraper.github import gov_orgs
from scraper import github, gitlab, bitbucket, doecode, tfs

logger = logging.getLogger(__name__)


def process_config(config):
    """
    Master function to process a Scraper config file

    Returns a Code.gov Metadata file
    """

    agency = config.get('agency', 'UNKNOWN')
    logger.debug('Agency: %s', agency)

    method = config.get('method', 'other')
    logger.debug('Inventory Method: %s', method)

    compute_labor_hours = config.get('compute_labor_hours', True)

    if config.get('contact_email', None) is None:
        # A default contact email is required to handle the (frequent) case
        # where a project / repository has no available contact email.
        logger.warning('Config file should contain a "contact_email"')

    logger.debug('Creating inventory from config: %s', config)
    code_gov_metadata = Metadata(agency, method)

    # Parse config for GitHub repositories
    github_instances = config.get('GitHub', [])
    if config.get('github_gov_orgs', False):
        github_instances.append({
            'url': 'https://github.com',
            'orgs': gov_orgs(),
        })
    for instance in github_instances:
        url = instance.get('url', 'https://github.com')
        orgs = instance.get('orgs', [])
        repos = instance.get('repos', [])
        public_only = instance.get('public_only', True)
        excluded = instance.get('exclude', [])
        token = instance.get('token', None)

        gh_session = github.connect(url, token)

        for repo in github.query_repos(gh_session, orgs, repos, public_only):
            if repo.owner.login in excluded or repo.full_name in excluded:
                logger.info('Excluding: %s', repo.full_name)
                continue

            code_gov_project = Project.from_github3(repo, labor_hours=compute_labor_hours)
            code_gov_metadata['releases'].append(code_gov_project)

    # Parse config for GitLab repositories
    gitlab_instances = config.get('GitLab', [])
    for instance in gitlab_instances:
        url = instance.get('url')
        # orgs = instance.get('orgs', [])
        repos = instance.get('repos', [])
        # public_only = instance.get('public_only', True)
        excluded = instance.get('exclude', [])
        token = instance.get('token', None)

        gl_session = gitlab.connect(url, token)

        for repo in gitlab.query_repos(gl_session, repos):
            namespace = repo.namespace['path']
            path_with_namespace = repo.path_with_namespace
            if namespace in excluded or path_with_namespace in excluded:
                logger.info('Excluding: %s', repo.path_with_namespace)
                continue

            code_gov_project = Project.from_gitlab(repo, labor_hours=compute_labor_hours)
            code_gov_metadata['releases'].append(code_gov_project)

    # Parse config for Bitbucket repositories
    bitbucket_instances = config.get('Bitbucket', [])
    for instance in bitbucket_instances:
        url = instance.get('url')
        # orgs = instance.get('orgs', None)
        # public_only = instance.get('public_only', True)
        # token = instance.get('token', None)
        username = instance.get('username')
        password = instance.get('password')
        excluded = instance.get('exclude', [])

        bb_session = bitbucket.connect(url, username, password)

        for repo in bitbucket.all_repos(bb_session):
            project = repo['project']['key']
            project_repo = '%s/%s' % (project, repo['slug'])
            if project in excluded or project_repo in excluded:
                logger.info('Excluding: %s', project_repo)
                continue

            code_gov_project = Project.from_stashy(repo, labor_hours=compute_labor_hours)
            code_gov_metadata['releases'].append(code_gov_project)

    # Parse config for TFS repositories
    tfs_instances = config.get('TFS', [])
    for instance in tfs_instances:
        url = instance.get('url')
        token = instance.get('token', None)

        projects = tfs.get_projects_metadata(url, token)
        for project in projects:
            code_gov_project = Project.from_tfs(project, labor_hours=compute_labor_hours)
            code_gov_metadata['releases'].append(code_gov_project)

    # Handle parsing of DOE CODE records

    doecode_config = config.get('DOE CODE', {})
    doecode_json = doecode_config.get('json', None)
    doecode_url = doecode_config.get('url', None)
    doecode_key = doecode_config.get('api_key', None)

    for record in doecode.process(doecode_json, doecode_url, doecode_key):
        code_gov_project = Project.from_doecode(record)
        code_gov_metadata['releases'].append(code_gov_project)

    return code_gov_metadata


def force_attributes(metadata, config):
    """
    Forces certain fields in the Code.gov Metadata json
    """

    organization = config.get('organization', '')
    logger.debug('Organization: %s', organization)

    contact_email = config.get('contact_email')
    logger.debug('Contact Email: %s', contact_email)

    permissions = config.get('permissions', {})
    default_usage = permissions.get('usageType', '')
    default_exemption_text = permissions.get('exemptionText', '')
    logger.debug('Default usageType: %s', default_usage)
    logger.debug('Default exemptionText: %s', default_exemption_text)

    # Force certain fields
    if organization:
        logger.debug('Forcing Organization to: %s', organization)

    if contact_email:
        logger.debug('Forcing Contact Email to: %s', contact_email)

    for release in metadata['releases']:
        if organization:
            release['organization'] = organization

        if contact_email:
            release['contact']['email'] = contact_email

        if 'licenses' not in release['permissions']:
            release['permissions']['licenses'] = None

        if 'description' not in release:
            release['description'] = 'No description available...'

        if 'usageType' not in release['permissions']:
            release['permissions']['usageType'] = default_usage
            release['permissions']['exemptionText'] = default_exemption_text

    return metadata



import logging
import os

import gitlab

logger = logging.getLogger(__name__)


def connect(url='https://gitlab.com', token=None):
    """
    Return a connected GitLab session

    ``token`` should be a ``private_token`` from Gitlab
    """

    if token is None:
        token = os.environ.get('GITLAB_API_TOKEN', None)

    gl_session = gitlab.Gitlab(url, token)

    try:
        gl_session.version()
    except (gitlab.execeptions.GitlabAuthenticationError):
        raise RuntimeError('Invalid or missing GITLAB_API_TOKEN')

    logger.info('Connected to: %s', url)

    return gl_session


def query_repos(gl_session, repos=None):
    """
    Yields Gitlab project objects for all projects in Bitbucket
    """

    if repos is None:
        repos = []

    for repo in repos:
        yield gl_session.projects.get(repo)

    if not repos:
        for project in gl_session.projects.list(as_list=False):
            yield project



import logging
import stashy

logger = logging.getLogger(__name__)


def connect(url, username, password):
    """
    Return a connected Bitbucket session
    """

    bb_session = stashy.connect(url, username, password)

    logger.info('Connected to: %s as %s', url, username)

    return bb_session


def all_repos(bb_session):
    """
    Yields Stashy repo objects for all projects in Bitbucket
    """

    for repo in bb_session.repos.all():
        yield repo



#! /usr/bin/env python3

import argparse
import json
import logging

from scraper.util import compute_labor_hours, git_repo_to_sloc


parser = argparse.ArgumentParser(description='Scrape code repositories for Code.gov / DOECode')
parser.add_argument('filename', type=str, help='Path to locally stored `code.json` file')
args = parser.parse_args()

code_gov_json = json.load(open(args.filename))
releases = code_gov_json['releases']

repo_urls = {
    release['repositoryURL'].rstrip('/')
    for release in releases
    if release.get('vcs', '') == 'git'
}

for url in repo_urls:
    # print(url)

    sloc = git_repo_to_sloc(url)
    # print(sloc)

    hours = compute_labor_hours(sloc)
    print('-- url=%s, sloc=%d, hours=%d' % (url, sloc, hours))



"""
A module for GitHub query and data management.

With this module, you will be able to send GraphQL and REST queries
to GitHub, as well as read and write JSON files to store data.
"""

import json
import os
import re
import time

from datetime import datetime
from subprocess import check_output, DEVNULL  # nosec

import pytz


def _vPrint(verbose, *args, **kwargs):
    """Easy verbosity-control print method.

    Args:
        verbose (bool): Normal print if True, do nothing otherwise.
        *args: Argument list for the 'print' method.
        **kwargs: Keyword arguments for the 'print' method.

    """
    if verbose:
        return print(*args, **kwargs)
    else:
        pass


class GitHubQueryManager:
    """GitHub query API manager."""

    def __init__(self, apiToken=None, maxRetry=10):
        """Initialize the GitHubQueryManager object.

        Note:
            If no apiToken argument is provided,
            the environment variable 'GITHUB_API_TOKEN' must be set.

        Args:
            apiToken (Optional[str]): A string representing a GitHub API
                token. Defaults to None.
            maxRetry (Optional[int]): A limit on how many times to
                automatically retry requests. Defaults to 10.

        Raises:
            TypeError: If no GitHub API token is provided either via
            argument or environment variable 'GITHUB_API_TOKEN'.

        """

        # Get GitHub API token
        if apiToken:
            self.__githubApiToken = apiToken
        else:
            try:
                self.__githubApiToken = os.environ['GITHUB_API_TOKEN']
            except KeyError as error:
                raise TypeError("Requires either a string argument or environment variable 'GITHUB_API_TOKEN'.") from error

        # Check token validity
        print("Checking GitHub API token... ", end="", flush=True)
        basicCheck = self._submitQuery('query { viewer { login } }')
        if basicCheck["statusNum"] == 401:
            print("FAILED.")
            raise ValueError("GitHub API token is not valid.\n" + basicCheck["headDict"]["http"] + " " + basicCheck["result"])
        else:
            print("Token validated.")

        # Initialize private variables
        self.__retryDelay = 3  #: Number of seconds to wait between retries.
        self.__query = None  #: Cached query string
        self.__queryPath = None  #: Path to query file
        self.__queryTimestamp = None  #: When query file was last modified

        # Initialize public variables
        self.maxRetry = maxRetry
        self.data = {}
        """Dict: Working data."""

    @property
    def maxRetry(self):
        """int: A limit on how many times to automatically retry requests.

        Must be a whole integer greater than 0.
        """
        return self.__maxRetry

    @maxRetry.setter
    def maxRetry(self, maxRetry):
        numIn = int(maxRetry)
        if not numIn > 0:
            numIn = 1
        self.__maxRetry = numIn
        print("Auto-retry limit for requests set to %d." % (self.maxRetry))

    def _readGQL(self, filePath, verbose=False):
        """Read a 'pretty' formatted GraphQL query file into a one-line string.

        Removes line breaks and comments. Condenses white space.

        Args:
            filePath (str): A relative or absolute path to a file containing
                a GraphQL query.
                File may use comments and multi-line formatting.
                .. _GitHub GraphQL Explorer:
                   https://developer.github.com/v4/explorer/
            verbose (Optional[bool]): If False, prints will be suppressed.
                Defaults to False.

        Returns:
            str: A single line GraphQL query.

        """
        if not os.path.isfile(filePath):
            raise RuntimeError("Query file '%s' does not exist." % (filePath))
        lastModified = os.path.getmtime(filePath)
        absPath = os.path.abspath(filePath)
        if absPath == self.__queryPath and lastModified == self.__queryTimestamp:
            _vPrint(verbose, "Using cached query '%s'" % (os.path.basename(self.__queryPath)))
            query_in = self.__query
        else:
            _vPrint(verbose, "Reading '%s' ... " % (filePath), end="", flush=True)
            with open(filePath, "r") as q:
                # Strip all comments and newlines.
                query_in = re.sub(r'#.*(\n|\Z)', '\n', q.read())
                # Condense extra whitespace.
                query_in = re.sub(r'\s+', ' ', query_in)
                # Remove any leading or trailing whitespace.
                query_in = re.sub(r'(\A\s+)|(\s+\Z)', '', query_in)
            _vPrint(verbose, "File read!")
            self.__queryPath = absPath
            self.__queryTimestamp = lastModified
            self.__query = query_in
        return query_in

    def queryGitHubFromFile(self, filePath, gitvars={}, verbosity=0, **kwargs):
        """Submit a GitHub GraphQL query from a file.

        Can only be used with GraphQL queries.
        For REST queries, see the 'queryGitHub' method.

        Args:
            filePath (str): A relative or absolute path to a file containing
                a GraphQL query.
                File may use comments and multi-line formatting.
                .. _GitHub GraphQL Explorer:
                   https://developer.github.com/v4/explorer/
            gitvars (Optional[Dict]): All query variables.
                Defaults to empty.
                GraphQL Only.
            verbosity (Optional[int]): Changes output verbosity levels.
                If < 0, all extra printouts are suppressed.
                If == 0, normal print statements are displayed.
                If > 0, additional status print statements are displayed.
                Defaults to 0.
            **kwargs: Keyword arguments for the 'queryGitHub' method.

        Returns:
            Dict: A JSON style dictionary.

        """
        gitquery = self._readGQL(filePath, verbose=(verbosity >= 0))
        return self.queryGitHub(gitquery, gitvars=gitvars, verbosity=verbosity, **kwargs)

    def queryGitHub(self, gitquery, gitvars={}, verbosity=0, paginate=False, cursorVar=None, keysToList=[], rest=False, requestCount=0, pageNum=0):
        """Submit a GitHub query.

        Args:
            gitquery (str): The query or endpoint itself.
                Examples:
                       query: 'query { viewer { login } }'
                    endpoint: '/user'
            gitvars (Optional[Dict]): All query variables.
                Defaults to empty.
                GraphQL Only.
            verbosity (Optional[int]): Changes output verbosity levels.
                If < 0, all extra printouts are suppressed.
                If == 0, normal print statements are displayed.
                If > 0, additional status print statements are displayed.
                Defaults to 0.
            paginate (Optional[bool]): Pagination will be completed
                automatically if True. Defaults to False.
            cursorVar (Optional[str]): Key in 'gitvars' that represents the
                pagination cursor. Defaults to None.
                GraphQL Only.
            keysToList (Optional[List[str]]): Ordered list of keys needed to
                retrieve the list in the query results to be extended by
                pagination. Defaults to empty.
                Example:
                    ['data', 'viewer', 'repositories', 'nodes']
                GraphQL Only.
            rest (Optional[bool]): If True, uses the REST API instead
                of GraphQL. Defaults to False.
            requestCount (Optional[int]): Counter for repeated requests.
            pageNum (Optional[int]): Counter for pagination.
                For user readable log messages only, does not affect data.

        Returns:
            Dict: A JSON style dictionary.

        """
        requestCount += 1
        if pageNum < 0:  # no negative page numbers
            pageNum = 0
        pageNum += 1

        if paginate:
            _vPrint((verbosity >= 0), "Page %d" % (pageNum))
        _vPrint((verbosity >= 0), "Sending %s query..." % ("REST" if rest else "GraphQL"))
        response = self._submitQuery(gitquery, gitvars=gitvars, verbose=(verbosity > 0), rest=rest)
        _vPrint((verbosity >= 0), "Checking response...")
        _vPrint((verbosity >= 0), response["headDict"]["http"])
        statusNum = response["statusNum"]

        # Decrement page count before error checks to properly reflect any repeated queries
        pageNum -= 1

        # Make sure the query limit didn't run out
        try:
            apiStatus = {
                "limit": int(response["headDict"]["X-RateLimit-Limit"]),
                "remaining": int(response["headDict"]["X-RateLimit-Remaining"]),
                "reset": int(response["headDict"]["X-RateLimit-Reset"])
            }
            _vPrint((verbosity >= 0), "API Status %s" % (json.dumps(apiStatus)))
            if not apiStatus["remaining"] > 0:
                _vPrint((verbosity >= 0), "API usage limit reached during query.")
                self._awaitReset(apiStatus["reset"])
                _vPrint((verbosity >= 0), "Repeating query...")
                return self.queryGitHub(gitquery, gitvars=gitvars, verbosity=verbosity, paginate=paginate, cursorVar=cursorVar, keysToList=keysToList, rest=rest, requestCount=(requestCount - 1), pageNum=pageNum)
        except KeyError:
            # Handles error cases that don't return X-RateLimit data
            _vPrint((verbosity >= 0), "Failed to check API Status.")

        # Check for accepted but not yet processed, usually due to un-cached data
        if statusNum == 202:
            if requestCount >= self.maxRetry:
                raise RuntimeError("Query attempted but failed %d times.\n%s\n%s" % (self.maxRetry, response["headDict"]["http"], response["result"]))
            else:
                self._countdown(self.__retryDelay, printString="Query accepted but not yet processed. Trying again in %*dsec...", verbose=(verbosity >= 0))
                return self.queryGitHub(gitquery, gitvars=gitvars, verbosity=verbosity, paginate=paginate, cursorVar=cursorVar, keysToList=keysToList, rest=rest, requestCount=requestCount, pageNum=pageNum)
        # Check for server error responses
        if statusNum == 502 or statusNum == 503:
            if requestCount >= self.maxRetry:
                raise RuntimeError("Query attempted but failed %d times.\n%s\n%s" % (self.maxRetry, response["headDict"]["http"], response["result"]))
            else:
                self._countdown(self.__retryDelay, printString="Server error. Trying again in %*dsec...", verbose=(verbosity >= 0))
                return self.queryGitHub(gitquery, gitvars=gitvars, verbosity=verbosity, paginate=paginate, cursorVar=cursorVar, keysToList=keysToList, rest=rest, requestCount=requestCount, pageNum=pageNum)
        # Check for other error responses
        if statusNum >= 400 or statusNum == 204:
            raise RuntimeError("Request got an Error response.\n%s\n%s" % (response["headDict"]["http"], response["result"]))

        _vPrint((verbosity >= 0), "Data received!")
        outObj = json.loads(response["result"])

        # Check for GraphQL API errors (e.g. repo not found)
        if not rest and "errors" in outObj:
            if requestCount >= self.maxRetry:
                raise RuntimeError("Query attempted but failed %d times.\n%s\n%s" % (self.maxRetry, response["headDict"]["http"], response["result"]))
            elif len(outObj["errors"]) == 1 and len(outObj["errors"][0]) == 1:
                # Poorly defined error type, usually intermittent, try again.
                _vPrint((verbosity >= 0), "GraphQL API error.\n%s" % (json.dumps(outObj["errors"])))
                self._countdown(self.__retryDelay, printString="Unknown API error. Trying again in %*dsec...", verbose=(verbosity >= 0))
                return self.queryGitHub(gitquery, gitvars=gitvars, verbosity=verbosity, paginate=paginate, cursorVar=cursorVar, keysToList=keysToList, rest=rest, requestCount=requestCount, pageNum=pageNum)
            else:
                raise RuntimeError("GraphQL API error.\n%s" % (json.dumps(outObj["errors"])))

        # Re-increment page count before the next page query
        pageNum += 1

        # Pagination
        if paginate:
            if rest and response["linkDict"]:
                if "next" in response["linkDict"]:
                    nextObj = self.queryGitHub(response["linkDict"]["next"], gitvars=gitvars, verbosity=verbosity, paginate=paginate, cursorVar=cursorVar, keysToList=keysToList, rest=rest, requestCount=0, pageNum=pageNum)
                    outObj.extend(nextObj)
            elif not rest:
                if not cursorVar:
                    raise ValueError("Must specify argument 'cursorVar' to use GraphQL auto-pagination.")
                if not len(keysToList) > 0:
                    raise ValueError("Must specify argument 'keysToList' as a non-empty list to use GraphQL auto-pagination.")
                aPage = outObj
                for key in keysToList[0:-1]:
                    aPage = aPage[key]
                gitvars[cursorVar] = aPage["pageInfo"]["endCursor"]
                if aPage["pageInfo"]["hasNextPage"]:
                    nextObj = self.queryGitHub(gitquery, gitvars=gitvars, verbosity=verbosity, paginate=paginate, cursorVar=cursorVar, keysToList=keysToList, rest=rest, requestCount=0, pageNum=pageNum)
                    newPage = nextObj
                    for key in keysToList[0:-1]:
                        newPage = newPage[key]
                    aPage[keysToList[-1]].extend(newPage[keysToList[-1]])
                aPage.pop("pageInfo", None)

        return outObj

    def _submitQuery(self, gitquery, gitvars={}, verbose=False, rest=False):
        """Send a curl request to GitHub.

        Args:
            gitquery (str): The query or endpoint itself.
                Examples:
                       query: 'query { viewer { login } }'
                    endpoint: '/user'
            gitvars (Optional[Dict]): All query variables.
                Defaults to empty.
            verbose (Optional[bool]): If False, stderr prints will be
                suppressed. Defaults to False.
            rest (Optional[bool]): If True, uses the REST API instead
                of GraphQL. Defaults to False.

        Returns:
            {
                'statusNum' (int): The HTTP status code.
                'headDict' (Dict[str]): The response headers.
                'linkDict' (Dict[int]): Link based pagination data.
                'result' (str): The body of the response.
            }

        """
        errOut = DEVNULL if not verbose else None
        authhead = 'Authorization: bearer ' + self.__githubApiToken

        bashcurl = 'curl -iH TMPauthhead -X POST -d TMPgitquery https://api.github.com/graphql' if not rest \
            else 'curl -iH TMPauthhead https://api.github.com' + gitquery
        bashcurl_list = bashcurl.split()
        bashcurl_list[2] = authhead
        if not rest:
            gitqueryJSON = json.dumps({'query': gitquery, 'variables': json.dumps(gitvars)})
            bashcurl_list[6] = gitqueryJSON

        fullResponse = check_output(bashcurl_list, stderr=errOut).decode()
        _vPrint(verbose, "\n" + fullResponse)
        fullResponse = fullResponse.split('\r\n\r\n')
        heads = fullResponse[0].split('\r\n')
        if len(fullResponse) > 1:
            result = fullResponse[1]
        else:
            result = ""
        http = heads[0].split()
        statusNum = int(http[1])

        # Parse headers into a useful dictionary
        headDict = {}
        headDict["http"] = heads[0]
        for header in heads[1:]:
            h = header.split(': ')
            headDict[h[0]] = h[1]

        # Parse any Link headers even further
        linkDict = None
        if "Link" in headDict:
            linkProperties = headDict["Link"].split(', ')
            propDict = {}
            for item in linkProperties:
                divided = re.split(r'<https://api.github.com|>; rel="|"', item)
                propDict[divided[2]] = divided[1]
            linkDict = propDict

        return {'statusNum': statusNum, 'headDict': headDict, 'linkDict': linkDict, 'result': result}

    def _awaitReset(self, utcTimeStamp, verbose=True):
        """Wait until the given UTC timestamp.

        Args:
            utcTimeStamp (int): A UTC format timestamp.
            verbose (Optional[bool]): If False, all extra printouts will be
                suppressed. Defaults to True.

        """
        resetTime = pytz.utc.localize(datetime.utcfromtimestamp(utcTimeStamp))
        _vPrint(verbose, "--- Current Timestamp")
        _vPrint(verbose, "      %s" % (time.strftime('%c')))
        now = pytz.utc.localize(datetime.utcnow())
        waitTime = round((resetTime - now).total_seconds()) + 1
        _vPrint(verbose, "--- Current UTC Timestamp")
        _vPrint(verbose, "      %s" % (now.strftime('%c')))
        _vPrint(verbose, "--- GITHUB NEEDS A BREAK Until UTC Timestamp")
        _vPrint(verbose, "      %s" % (resetTime.strftime('%c')))
        self._countdown(waitTime, printString="--- Waiting %*d seconds...", verbose=verbose)
        _vPrint(verbose, "--- READY!")

    def _countdown(self, waitTime=0, printString="Waiting %*d seconds...", verbose=True):
        """Makes a pretty countdown.

        Args:
            gitquery (str): The query or endpoint itself.
                Examples:
                       query: 'query { viewer { login } }'
                    endpoint: '/user'
            printString (Optional[str]): A counter message to display.
                Defaults to 'Waiting %*d seconds...'
            verbose (Optional[bool]): If False, all extra printouts will be
                suppressed. Defaults to True.

        """
        if waitTime <= 0:
            waitTime = self.__retryDelay
        for remaining in range(waitTime, 0, -1):
            _vPrint(verbose, "\r" + printString % (len(str(waitTime)), remaining), end="", flush=True)
            time.sleep(1)
        if verbose:
            _vPrint(verbose, "\r" + printString % (len(str(waitTime)), 0))


class DataManager:
    """JSON data manager."""

    def __init__(self, filePath=None, loadData=False):
        """Initialize the DataManager object.
        Args:
            filePath (Optional[str]): Relative or absolute path to a JSON
                data file. Defaults to None.
            loadData (Optional[bool]): Loads data from the given file path
                if True. Defaults to False.

        """
        self.data = {}
        """Dict: Working data."""
        self.filePath = filePath
        if loadData:
            self.fileLoad(updatePath=False)

    @property
    def filePath(self):
        """str: Absolute path to a JSON format data file.

        Can accept relative paths, but will always convert them to
        the absolute path.
        """
        if not self.__filePath:
            raise ValueError("Internal variable filePath has not been set.")
        return self.__filePath

    @filePath.setter
    def filePath(self, filePath):
        if filePath:
            if not os.path.isfile(filePath):
                print("Data file '%s' does not currently exist. Saving data will create a new file." % (filePath))
            self.__filePath = os.path.abspath(filePath)
            print("Stored new data file path '%s'" % (self.filePath))
        else:
            self.__filePath = None

    def dataReset(self):
        """Reset the internal JSON data dictionary."""
        self.data = {}
        print("Stored data has been reset.")

    def fileLoad(self, filePath=None, updatePath=True):
        """Load a JSON data file into the internal JSON data dictionary.

        Current internal data will be overwritten.
        If no file path is provided, the stored data file path will be used.

        Args:
            filePath (Optional[str]): A relative or absolute path to a
                '.json' file. Defaults to None.
            updatePath (Optional[bool]): Specifies whether or not to update
                the stored data file path. Defaults to True.

        """
        if not filePath:
            filePath = self.filePath
        if not os.path.isfile(filePath):
            raise FileNotFoundError("Data file '%s' does not exist." % (filePath))
        else:
            print("Importing existing data file '%s' ... " % (filePath), end="", flush=True)
            with open(filePath, "r") as q:
                data_raw = q.read()
            print("Imported!")
            self.data = json.loads(data_raw)
            if updatePath:
                self.filePath = filePath

    def fileSave(self, filePath=None, updatePath=False):
        """Write the internal JSON data dictionary to a JSON data file.

        If no file path is provided, the stored data file path will be used.

        Args:
            filePath (Optional[str]): A relative or absolute path to a
                '.json' file. Defaults to None.
            updatePath (Optional[bool]): Specifies whether or not to update
                the stored data file path. Defaults to False.

        """
        if not filePath:
            filePath = self.filePath
        if not os.path.isfile(filePath):
            print("Data file '%s' does not exist, will create new file." % (filePath))
            if not os.path.exists(os.path.split(filePath)[0]):
                os.makedirs(os.path.split(filePath)[0])
        dataJsonString = json.dumps(self.data, indent=4, sort_keys=True)
        print("Writing to file '%s' ... " % (filePath), end="", flush=True)
        with open(filePath, "w") as fileout:
            fileout.write(dataJsonString)
        print("Wrote file!")
        if updatePath:
            self.filePath = filePath



import logging

logger = logging.getLogger(__name__)


def _license_obj(license):
    """
    A helper function to look up license object information

    Use names from: https://api.github.com/licenses
    """
    obj = None

    if license in ('MIT', 'MIT License'):
        obj = {
            'URL': 'https://api.github.com/licenses/mit',
            'name': 'MIT'
        }
    elif license in ('BSD 2-clause "Simplified" License'):
        obj = {
            'URL': 'https://api.github.com/licenses/bsd-2-clause',
            'name': 'BSD-2-Clause'
        }
    elif license in ('BSD 3-clause "New" or "Revised" License'):
        obj = {
            'URL': 'https://api.github.com/licenses/bsd-3-clause',
            'name': 'BSD-3-Clause'
        }
    elif license in ('Apache License 2.0'):
        obj = {
            'URL': 'https://api.github.com/licenses/apache-2.0',
            'name': 'Apache-2.0'
        }
    elif license in ('GNU General Public License v2.1'):
        obj = {
            'URL': 'https://api.github.com/licenses/gpl-2.1',
            'name': 'GPL-2.1'
        }
    elif license in ('GNU General Public License v2.0'):
        obj = {
            'URL': 'https://api.github.com/licenses/gpl-2.0',
            'name': 'GPL-2.0'
        }
    elif license in ('GNU Lesser General Public License v2.1'):
        obj = {
            'URL': 'https://api.github.com/licenses/lgpl-2.1',
            'name': 'LGPL-2.1'
        }
    elif license in ('GNU General Public License v3.0'):
        obj = {
            'URL': 'https://api.github.com/licenses/gpl-3.0',
            'name': 'GPL-3.0'
        }
    elif license in ('GNU Lesser General Public License v3.0'):
        obj = {
            'URL': 'https://api.github.com/licenses/lgpl-3.0',
            'name': 'LGPL-3.0'
        }
    elif license in ('Eclipse Public License 1.0'):
        obj = {
            'URL': 'https://api.github.com/licenses/epl-1.0',
            'name': 'EPL-1.0',
        }
    elif license in ('Mozilla Public License 2.0'):
        obj = {
            'URL': 'https://api.github.com/licenses/mpl-2.0',
            'name': 'MPL-2.0',
        }
    elif license in ('The Unlicense'):
        obj = {
            'URL': 'https://api.github.com/licenses/unlicense',
            'name': 'Unlicense',
        }
    elif license in ('GNU Affero General Public License v3.0'):
        obj = {
            'URL': 'https://api.github.com/licenses/agpl-3.0',
            'name': 'AGPL-3.0',
        }
    elif license in ('Eclipse Public License 2.0'):
        obj = {
            'URL': 'https://api.github.com/licenses/epl-2.0',
            'name': 'EPL-2.0',
        }

    if obj is None:
        logger.warn('I dont understand the license: %s', license)
        raise ValueError('Aborting!')

    return obj



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.md') as fh:
    long_description = fh.read()

with open('requirements/production.txt') as fp:
    lines = [x.strip() for x in fp.readlines() if x]
    install_reqs = [x for x in lines if not x.startswith('#')]

setup(
    name='llnl-scraper',
    version='0.7.0',
    description='Package for extracting software repository metadata',
    long_description=long_description,
    author='Ian Lee',
    author_email='lee1001@llnl.gov',
    url='https://github.com/llnl/scraper',
    packages=find_packages(),
    install_requires=install_reqs,
    entry_points={
        'console_scripts': [
            'scraper = scraper.gen_code_gov_json:main',
        ]
    },
    scripts=[
        'scripts/codegov_compute_hours.py',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
)



class TFSProject:
    def __init__(self, projectInfo, collectionInfo):
        self.projectInfo = projectInfo
        self.collectionInfo = collectionInfo
        self.projectCreateInfo = {}
        self.projectLastUpdateInfo = {}
        self.gitInfo = []
        self.tfvcInfo = []



#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import logging
import os

import requests
import github3
import time

logger = logging.getLogger(__name__)


def gov_orgs():
    """
    Returns a list of the names of US Government GitHub organizations

    Based on: https://government.github.com/community/

    Exmample return:
        {'llnl', '18f', 'gsa', 'dhs-ncats', 'spack', ...}
    """
    us_gov_github_orgs = set()

    gov_orgs = requests.get('https://government.github.com/organizations.json').json()

    us_gov_github_orgs.update(gov_orgs['governments']['U.S. Federal'])
    us_gov_github_orgs.update(gov_orgs['governments']['U.S. Military and Intelligence'])
    us_gov_github_orgs.update(gov_orgs['research']['U.S. Research Labs'])

    return list(us_gov_github_orgs)


def create_session(token=None):
    """
    Create a github3.py session connected to GitHub.com

    If token is not provided, will attempt to use the GITHUB_API_TOKEN
    environment variable if present.
    """
    if token is None:
        token = os.environ.get('GITHUB_API_TOKEN', None)

    gh_session = github3.login(token=token)

    if gh_session is None:
        raise RuntimeError('Invalid or missing GITHUB_API_TOKEN')

    return gh_session


def create_enterprise_session(url, token=None):
    """
    Create a github3.py session for a GitHub Enterprise instance

    If token is not provided, will attempt to use the GITHUB_API_TOKEN
    environment variable if present.
    """

    gh_session = github3.enterprise_login(url=url, token=token)

    if gh_session is None:
        msg = 'Unable to connect to GitHub Enterprise (%s) with provided token.'
        raise RuntimeError(msg, url)

    return gh_session


def _num_requests_needed(num_repos, factor=2, wiggle_room=100):
    """
    Helper function to estimate the minimum number of API requests needed
    """
    return num_repos * factor + wiggle_room


def _check_api_limits(gh_session, api_required=250, sleep_time=15):
    """
    Simplified check for API limits

    If necessary, spin in place waiting for API to reset before returning.

    See: https://developer.github.com/v3/#rate-limiting
    """
    api_rates = gh_session.rate_limit()

    api_remaining = api_rates['rate']['remaining']
    api_reset = api_rates['rate']['reset']
    logger.debug('Rate Limit - %d requests remaining', api_remaining)

    if api_remaining > api_required:
        return

    now_time = time.time()
    time_to_reset = int(api_reset - now_time)
    logger.warn('Rate Limit Depleted - Sleeping for %d seconds', time_to_reset)

    while now_time < api_reset:
        time.sleep(10)
        now_time = time.time()

    return


def connect(url='https://github.com', token=None):
    """
    Create a GitHub session for making requests
    """

    gh_session = None
    if url == 'https://github.com':
        gh_session = create_session(token)
    else:
        gh_session = create_enterprise_session(url, token)

    if gh_session is None:
        msg = 'Unable to connect to (%s) with provided token.'
        raise RuntimeError(msg, url)

    logger.info('Connected to: %s', url)

    return gh_session


def query_repos(gh_session, orgs=None, repos=None, public_only=True):
    """
    Yields GitHub3.py repo objects for provided orgs and repo names

    If orgs and repos are BOTH empty, execute special mode of getting ALL
    repositories from the GitHub Server.

    If public_only is True, will return only those repos that are marked as
    public. Set this to false to return all organizations that the session has
    permissions to access.
    """

    if orgs is None:
        orgs = []
    if repos is None:
        repos = []
    if public_only:
        privacy = 'public'
    else:
        privacy = 'all'

    _check_api_limits(gh_session, 10)

    for org_name in orgs:
        org = gh_session.organization(org_name)
        num_repos = org.public_repos_count

        _check_api_limits(gh_session, _num_requests_needed(num_repos))

        for repo in org.repositories(type=privacy):
            _check_api_limits(gh_session, 10)
            yield repo

    for repo_name in repos:
        _check_api_limits(gh_session, 10)
        org, name = repo_name.split('/')
        yield gh_session.repository(org, name)

    if not (orgs or repos):
        for repo in gh_session.all_repositories():
            yield repo
